unit JarBook;

interface

uses
  Windows, SysUtils, Classes, Contnrs, ztvUnJar, ztvJar, ztvBase,
  ztvStreams, ztvZip, BookType;

type
  TBookChapter = class
  private
    Stream: TMemoryStream32;
    Text: TStringStream;
    function GetText: string;
    procedure SetText(const Value: string);
    procedure CopyFrom(Source: TStream32; Count: Int64);
    function SaveText(FileName: string): Integer;
    function SaveJar(FileName: string): Integer;
  public
    { ļ }
    FileName: string;
    { ڲ }
    InnerName: string;

    constructor Create;
    destructor Destroy; override;
    { ı }
    property AsText: string read GetText write SetText;
  end;

  TJarBook = class
  private
    FFileName: string;
    UnJar: TUnJar;
    DecompMemoryStream: TDecompMemoryStream;
    Jad, MANIFEST: TStrings;
    FSavedTextFileName: string;
    FSavedJarFileName: string;
    function GetChapter(Name: string): TBookChapter;
    function GetIsValid: Boolean;
    procedure ClearChapters;
    function AddChapter(InnerName, FileName: string; Content: string = ''): Integer; overload;
    function AddChapter(Info: string): Integer; overload;
    procedure SaveBookInfo(FileName: string);
    procedure UpdateConfig(Config: TStrings);
  public
    { ļϢ }
    BookInfo: TBookInfo;
    ChapterList: TObjectList;

    constructor Create; overload;
    constructor Create(FileName: string); overload;
    destructor Destroy; override;

    { Ϣ }
    procedure LoadBook(FileName: string; FullLoad: Boolean = True);
    { Ϊıʽ }
    function SaveTextBook(FileName: string = ''): Integer;
    { ΪJarʽ }
    function SaveJarBook(FileName: string = ''; OnlyInfo: Boolean = False): Integer;
    { ļ }
    property FileName: string read FFileName write FFileName;
    { ½ }
    property Chapters[Name: string]: TBookChapter read GetChapter;
    { ʽǷЧ }
    property IsValid: Boolean read GetIsValid;
    { ıļ }
    property SavedTextFileName: string read FSavedTextFileName;
    { Jarļ }
    property SavedJarFileName: string read FSavedJarFileName;
  end;

implementation

uses
  Utils;

const
  BookSymbol = 'HB';
  ConfigFile = 'META-INF\MANIFEST.MF';

{ TJarBook }

function TJarBook.AddChapter(InnerName, FileName, Content: string): Integer;
var
  bc: TBookChapter;
begin
  bc := TBookChapter.Create;
  bc.InnerName := InnerName;
  bc.FileName := FileName;
  if Content <> '' then
    bc.AsText := Content;
  ChapterList.Add(bc);
  Result := ChapterList.Count - 1;
end;

function TJarBook.AddChapter(Info: string): Integer;
var
  i, j: Integer;
  s1, s2, s3: string;
begin
  j := 0;
  s1 := '';
  s2 := '';
  if Pos(',', Info) > 0 then
    for i := 1 to Length(Info) do
    begin
      if (Info[i] = ',') and (j < 2) then
        Inc(j)
      else
        case j of
          0:
            s1 := s1 + Info[i];
          1:
            s2 := s2 + Info[i];
        else
          s3 := s3 + Info[i];
        end;
    end;
  if (s1 <> '') and (s2 <> '') and (s3 <> '') then
    Result := AddChapter(s1, s3)
  else
    Result := -1;
end;

procedure TJarBook.ClearChapters;
var
  i: Integer;
begin
  for i := 0 to ChapterList.Count - 1 do
    (ChapterList[i] as TBookChapter).Free;
  ChapterList.Clear;
end;

constructor TJarBook.Create;
begin
  Jad := TStringList.Create;
  MANIFEST := TStringList.Create;
  ChapterList := TObjectList.Create(False);
  UnJar := TUnJar.Create(nil);
  DecompMemoryStream := TDecompMemoryStream.Create;
  FFileName := '';
  FSavedTextFileName := '';
  FSavedJarFileName := '';
end;

constructor TJarBook.Create(FileName: string);
begin
  Create;
  if FileExists(FileName) then
  begin
    LoadBook(FileName);
    FFileName := FileName;
  end;
end;

destructor TJarBook.Destroy;
begin
  DecompMemoryStream.ClearMemory;
  DecompMemoryStream.Free;
  UnJar.Free;
  ClearChapters;
  ChapterList.Free;
  Jad.Free;
  MANIFEST.Free;
  inherited Destroy;
end;

function TJarBook.GetChapter(Name: string): TBookChapter;
var
  i: Integer;
  bc: TBookChapter;
begin
  Result := nil;
  for i := 0 to ChapterList.Count - 1 do
  begin
    bc := ChapterList[i] as TBookChapter;
    if CompareText(bc.InnerName, Name) = 0 then
    begin
      Result := bc;
      Break;
    end;
  end;
end;

function TJarBook.GetIsValid: Boolean;
begin
  Result := (BookInfo.FileName <> '') and (BookInfo.Author <> '')
    and (BookInfo.Maker <> '') and (BookInfo.Contact <> '');
end;

procedure TJarBook.LoadBook(FileName: string; FullLoad: Boolean);
var
  Buf: array of Char;
  P: PChar;
  s: string;
  i: Integer;
  bc: TBookChapter;

  function GetString(Count: Integer): string;
  begin
    DecompMemoryStream.Read(Buf[0], Count);
    i := Ord(Buf[Count - 2]) shl 8 + Ord(Buf[Count - 1]);
    if (i > 0) and (DecompMemoryStream.Position < DecompMemoryStream.Size) then
    begin
      DecompMemoryStream.Read(Buf[0], i);
      Buf[i] := #0;
      Result := Trim(UTF8ToWideString(StrPas(P)));
    end
    else
      Result := '';
  end;

begin
  BookInfo.FilePath := '';
  BookInfo.FileName := '';
  BookInfo.BookName := '';
  BookInfo.Author := '';
  BookInfo.Maker := '';
  BookInfo.Contact := '';
  BookInfo.Size := 0;
  if FileExists(FileName) then
  begin
    FFileName := FileName;
    UnJar.ArchiveFile := FileName;
    UnJar.RecurseDirs := True;
    UnJar.ConfirmOverwrites := True;
    UnJar.FileSpec.Text := '0';
    DecompMemoryStream.ClearMemory;
    UnJar.ExtractToMemoryStream(DecompMemoryStream);
    if DecompMemoryStream.FileCount = 1 then
    begin
      SetLength(Buf, 2048);
      P := @Buf[0];
      DecompMemoryStream.Seek(0, soFromBeginning);
      BookInfo.FileName := GetString(5);
      ClearChapters;
      Jad.Clear;
      MANIFEST.Clear;
      UnJar.FileSpec.Clear;
      s := GetString(5);
      while True do
      begin
        if Length(s) = 0 then
          Break;
        if FullLoad then
        begin
          i := AddChapter(s);
          if i >= 0 then
            UnJar.FileSpec.Add((ChapterList[i] as TBookChapter).InnerName);
        end;
        s := GetString(2);
      end;
      BookInfo.BookName := GetString(2);
      BookInfo.Author := GetString(2);
      BookInfo.Maker := GetString(2);
      BookInfo.Contact := GetString(2);
      BookInfo.EditDate := GetFileDate(FileName);
      BookInfo.Size := UnJar.LengthOfFile;
      BookInfo.BookType := btJar;
      BookInfo.FilePath := ExpandFileName(FileName);
      if BookInfo.BookName = '' then
        BookInfo.BookName := BookInfo.FileName;

      DecompMemoryStream.ClearMemory;
      if FullLoad then
      begin
        UnJar.ExtractToMemoryStream(DecompMemoryStream);
        if DecompMemoryStream.FileCount > 0 then
          for i := 0 to DecompMemoryStream.FileCount - 1 do
          begin
            bc := Chapters[DecompMemoryStream.FileName[i]];
            if Assigned(bc) then
            begin
              DecompMemoryStream.Seek(DecompMemoryStream.FileOffset[i], soFromBeginning);
              bc.CopyFrom(DecompMemoryStream, DecompMemoryStream.FileSize[i]);
            end;
          end;
      end;

      UnJar.FileSpec.Clear;
      UnJar.FileSpec.Add(ConfigFile);
      DecompMemoryStream.ClearMemory;
      UnJar.ExtractToMemoryStream(DecompMemoryStream);
      if DecompMemoryStream.FileCount = 1 then
      begin
        DecompMemoryStream.Seek(DecompMemoryStream.FileOffset[0], soFromBeginning);
        MANIFEST.LoadFromStream(DecompMemoryStream);
      end;

      s := ChangeFileExt(FFileName, '.jad');
      if FileExists(s) then
        Jad.LoadFromFile(s);
    end;
  end;
end;

type
  TBackupMode = (bmCopy, bmMove);

procedure CreateBackupFile(Name: string; BackupMode: TBackupMode);
var
  BakFile: string;
begin
  if FileExists(Name) then
  begin
    BakFile := ChangeFileExt(Name, '.bak');
    if CompareText(Name, BakFile) <> 0 then
    begin
      if FileExists(BakFile) then
        DeleteFile(BakFile);
      if BackupMode = bmMove then
        RenameFile(Name, BakFile)
      else
        CopyFile(PChar(Name), PChar(BakFile), False);
    end;
  end;
end;

procedure TJarBook.SaveBookInfo(FileName: string);
var
  fs: TFileStream;
  j: Integer;
  bc: TBookChapter;

  procedure WriteString(s: string);
  var
    i: Integer;
    Buf: array[0..1] of Byte;
  begin
    if s <> '' then
      s := WideStringToUTF8(s);
    i := Length(s);
    Buf[0] := i div 256;
    Buf[1] := i mod 256;
    fs.WriteBuffer(Buf[0], 2);
    if s <> '' then
      fs.WriteBuffer(Pointer(s)^, Length(s));
  end;
begin
  if FileExists(FileName) then
    DeleteFile(FileName);
  fs := TFileStream.Create(FileName, fmCreate);
  try
    WriteString('0');
    WriteString(BookInfo.FileName);
    WriteString(IntToStr(ChapterList.Count));
    for j := 0 to ChapterList.Count - 1 do
    begin
      bc := ChapterList[j] as TBookChapter;
      WriteString(Format('%s,%d,%s', [bc.InnerName, Length(bc.AsText), bc.FileName]));
    end;
    WriteString('');
    WriteString(BookInfo.BookName);
    WriteString(BookInfo.Author);
    WriteString(BookInfo.Maker);
    WriteString(BookInfo.Contact);
  finally
    fs.Free;
  end;
end;

function TJarBook.SaveJarBook(FileName: string; OnlyInfo: Boolean): Integer;
  procedure OptimizeJar;
  var
    bc: TBookChapter;
    i: Integer;
  begin
    i := 0;
    while True do
    begin
      if (ChapterList.Count = 1) or (ChapterList.Count <= i) then
        Break;

      bc := ChapterList[i] as TBookChapter;
      if bc.Text.Size = 0 then
      begin
        ChapterList.Delete(i);
        bc.Free;
      end
      else
      begin
        bc.InnerName := IntToStr(i + 1);
        Inc(i);
      end;
    end;
  end;
var
  TempFile, TempPath, JadFile: string;
  Jar: TJar;
  i: Integer;
  bc: TBookChapter;
begin
  Result := -1;
  if not IsValid or not FileExists(FFileName) then
    Exit;
  if MANIFEST.Count <= 0 then
    Exit;

  if FileName = '' then
    FileName := FFileName;
  FileName := ExpandFileName(FileName);
  FileName := ChangeFileExt(FileName, '.jar');
  CreateBackupFile(FileName, bmCopy);
  if CompareText(ExpandFileName(FFileName), ExpandFileName(FileName)) <> 0 then
    CopyFile(PChar(FFileName), PChar(FileName), False);
  TempPath := GetTempDir + '~UnJar~';
  if not DirectoryExists(TempPath) then
  begin
    CreateDir(TempPath);
    CreateDir(AddBackSlash(TempPath) + 'META-INF');
  end;
  TempPath := AddBackSlash(TempPath);
  Jar := TJar.Create(nil);
  try
    Jar.ArchiveFile := FileName;
    Jar.DateAttribute := daFileDate;
    Jar.CompressMethod := cmDeflate;
    Jar.RecurseDirs := True;
    Jar.StoreEmptySubDirs := True;
    Jar.EncryptHeaders := False;
    Jar.Switch := swDelete;
    Jar.FileSpec.Add('0');
    Jar.FileSpec.Add(ConfigFile);
    if not OnlyInfo then
    begin
      for i := 0 to ChapterList.Count - 1 do
      begin
        bc := ChapterList[i] as TBookChapter;
        Jar.FileSpec.Add(bc.InnerName);
      end;
    end;
    if Jar.FileSpec.Count > 0 then
      Jar.Compress;

    Jar.Switch := swAdd;
    Jar.StoredDirNames := sdExplorer_UserDefineDefaultDir;
    Jar.DefaultDir := TempPath;

    if not OnlyInfo then
      OptimizeJar;

    Jar.FileSpec.Clear;
    TempFile := TempPath + '0';
    Jar.FileSpec.Add(TempFile);
    SaveBookInfo(TempFile);
    TempFile := TempPath + ConfigFile;
    UpdateConfig(MANIFEST);
    MANIFEST.SaveToFile(TempFile);
    Jar.FileSpec.Add(TempFile);
    if not OnlyInfo then
    begin
      for i := 0 to ChapterList.Count - 1 do
      begin
        bc := ChapterList[i] as TBookChapter;
        TempFile := TempPath + bc.InnerName;
        if bc.SaveJar(TempFile) >= 0 then
          Jar.FileSpec.Add(TempFile);
      end;
    end;
    Jar.Compress;

    if Jad.Count > 0 then
    begin
      UpdateConfig(Jad);
      JadFile := ChangeFileExt(FileName, '.jad');
      Jad.SaveToFile(JadFile);
    end;

    if ChangeFileExt(ExtractFileName(FileName), '') <> BookInfo.FileName then
    begin
      FSavedJarFileName := AddBackSlash(ExtractFilePath(FileName)) + BookInfo.FileName + '.jar';
      if not FileExists(FSavedJarFileName) then
      begin
        RenameFile(FileName, FSavedJarFileName);
        TempFile := ChangeFileExt(FSavedJarFileName, '.jad');
        if FileExists(TempFile) then
          DeleteFile(TempFile);
        RenameFile(JadFile, TempFile);
      end
      else
        FSavedJarFileName := FileName;
    end
    else
      FSavedJarFileName := FileName;
    Result := 0;
  finally
    if DirectoryExists(TempPath) then
      Deltree(TempPath);
    Jar.Free;
  end;
end;

function TJarBook.SaveTextBook(FileName: string): Integer;
var
  bc: TBookChapter;
  TempPath, TempFile: string;
  Zip: TZip;
  i: Integer;
begin
  Result := -1;
  if not IsValid then
    Exit;

  if FileName = '' then
    FileName := FFileName;
  FileName := ExpandFileName(FileName);
  FileName := ExtractFilePath(FileName) + BookInfo.FileName;
  if ChapterList.Count > 1 then
  begin
    FileName := ChangeFileExt(FileName, '.zip');
    CreateBackupFile(FileName, bmMove);
    TempPath := GetTempDir + '~UnJar~';
    if not DirectoryExists(TempPath) then
      CreateDir(TempPath);
    TempPath := AddBackSlash(TempPath);
    Zip := TZip.Create(nil);
    try
      Zip.ArchiveFile := FileName;
      Zip.DateAttribute := daFileDate;
      Zip.StoredDirNames := sdNone;
      Zip.CompressMethod := cmDeflate;
      Zip.RecurseDirs := True;
      Zip.Switch := swAdd;
      Zip.StoreEmptySubDirs := False;
      Zip.EncryptHeaders := False;
      for i := 0 to ChapterList.Count - 1 do
      begin
        bc := ChapterList[i] as TBookChapter;
        TempFile := TempPath + bc.FileName + '.txt';
        if bc.SaveText(TempFile) >= 0 then
          Zip.FileSpec.Add(TempFile);
      end;
      Zip.Compress;
      Result := 0;
    finally
      if DirectoryExists(TempPath) then
        Deltree(TempPath);
      Zip.Free;
    end;
  end
  else
  begin
    FileName := ChangeFileExt(FileName, '.txt');
    CreateBackupFile(FileName, bmMove);
    Result := (ChapterList[0] as TBookChapter).SaveText(FileName);
  end;
  FSavedTextFileName := FileName;
end;

procedure TJarBook.UpdateConfig(Config: TStrings);
const
  MIDlet_1 = 'MIDlet-1:';
  MIDlet_Jar_Size = 'MIDlet-Jar-Size:';
  MIDlet_Jar_URL = 'MIDlet-Jar-URL:';
  MIDlet_Name = 'MIDlet-Name:';
var
  i: Integer;
  BookName: string;
begin
  BookName := WideStringToUTF8(BookInfo.BookName);
  for i := 0 to Config.Count - 1 do
  begin
    Config[i] := Trim(Config[i]);
    if Pos(MIDlet_1, Config[i]) = 1 then
      Config[i] := MIDlet_1 + ' ' + BookName + ', , JavaBook'
    else if Pos(MIDlet_Jar_Size, Config[i]) = 1 then
      Config[i] := MIDlet_Jar_Size + ' ' + IntToStr(FileSize(FFileName))
    else if Pos(MIDlet_Jar_URL, Config[i]) = 1 then
      Config[i] := MIDlet_Jar_URL + ' ' + WideStringToUTF8(ExtractFileName(FFileName))
    else if Pos(MIDlet_Name, Config[i]) = 1 then
      Config[i] := MIDlet_Name + ' ' + BookName;
  end;
end;

{ TBookChapter }

procedure TBookChapter.CopyFrom(Source: TStream32; Count: Int64);
var
  Buf: array of Char;
begin
  SetLength(Buf, 2);
  Source.Read(Buf[0], 2);
  if BookSymbol = Buf[0] + Buf[1] then
    Stream.CopyFrom(Source, Count - 2)
  else
    Stream.Clear;
  Text.Size := 0;
end;

constructor TBookChapter.Create;
begin
  Stream := TMemoryStream32.Create;
  Text := TStringStream.Create('');
end;

destructor TBookChapter.Destroy;
begin
  Stream.Free;
  Text.Free;
  inherited Destroy;
end;

function TBookChapter.GetText: string;
var
  i, Count: Integer;
  Buf: array of Char;
begin
  if (Text.Size <= 0) and (Stream.Size > 0) then
  begin
    Stream.Position := 0;
    SetLength(Buf, 1024 * 512);
    while True do
    begin
      Count := Stream.Read(Buf[0], Length(Buf));
      for i := 0 to (Count div 2) - 1 do
      begin
        if Buf[i * 2 + 1] = #0 then
          Text.WriteString(Buf[i * 2])
        else
          Text.WriteString(WideChar(Ord(Buf[i * 2 + 1]) shl 8 + Ord(Buf[i * 2])));
      end;
      if Stream.Position >= Stream.Size - 1 then
        Break;
    end;
  end;
  Result := Text.DataString;
end;

function TBookChapter.SaveJar(FileName: string): Integer;
var
  fs: TFileStream;
  s: string;
begin
  try
    if FileExists(FileName) then
      DeleteFile(FileName);
    fs := TFileStream.Create(FileName, fmCreate);
    try
      s := BookSymbol;
      fs.WriteBuffer(Pointer(s)^, Length(s));
      fs.WriteBuffer(Stream.Memory^, Stream.Size);
      Result := 0;
    finally
      fs.Free;
    end;
  except
    Result := -1;
  end;
end;

function TBookChapter.SaveText(FileName: string): Integer;
var
  fs: TFileStream;
  s: string;
  ss: TStrings;
  i: Integer;
begin
  try
    if FileExists(FileName) then
      DeleteFile(FileName);
    fs := TFileStream.Create(FileName, fmCreate);
    ss := TStringList.Create;
    try
      ss.Text := AsText;
      for i := 0 to ss.Count - 1 do
        ss[i] := TrimRight(ss[i]);
      s := ss.Text;
      fs.WriteBuffer(Pointer(s)^, Length(s));
      Result := 0;
    finally
      fs.Free;
      ss.Free;
    end;
  except
    Result := -1;
  end;
end;

procedure TBookChapter.SetText(const Value: string);
var
  s: WideString;
  i: Integer;
  Buf: array of Char;
begin
  Stream.Clear;
  if Value <> '' then
  begin
    Text.Size := Length(Value);
    Text.Seek(0, soFromBeginning);
    Text.WriteString(Value);
    SetLength(Buf, 2);
    s := Text.DataString;
    for i := 1 to Length(s) do
    begin
      Buf[0] := Chr(Word(s[i]) and $00FF);
      Buf[1] := Chr(Word(s[i]) shr 8);
      Stream.Write(Buf[0], 2);
    end;
  end
  else
    Text.Size := 0;
end;

end.

